%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Definitions of builtins of module IOExts:
%

:- use_module('../prologbasics').
:- ensure_loaded(user:prim_ports). % for IOExts.choiceStreamOrMsgs

:- dynamic globalAssoc/2.

?- block 'prim_setAssoc'(?,?,?,?,-,?).
'prim_setAssoc'(RKey,RVal,_,'$io'('Prelude.()'),E0,E) :-
        derefAll(RKey,Key), derefAll(RVal,Val),
        string2Atom(Key,KeyA),
	(retract(globalAssoc(KeyA,_)) -> true ; true),
	assertz(globalAssoc(KeyA,Val)),
	!, E0=E.

?- block 'prim_getAssoc'(?,?,?,-,?).
'prim_getAssoc'(RKey,_,'$io'(R),E0,E) :-
        derefAll(RKey,Key),
        string2Atom(Key,KeyA),
	(globalAssoc(KeyA,Val) -> R='Prelude.Just'(Val) ; R='Prelude.Nothing'),
	!, E0=E.

% shell command execution:
?- block prim_execCmd(?,?,?,-,?).
prim_execCmd(RCmdString,_,'$io'('Prelude.(,,)'(StdIn,StdOut,StdErr)),E0,E) :-
        derefAll(RCmdString,CmdString),
	string2Atom(CmdString,Cmd),
	execCommand(Cmd,StdIn,StdOut,StdErr), E0=E.


% shell command execution:
?- block prim_connectToCmd(?,?,?,-,?).
prim_connectToCmd(RCmdString,_,
		  '$io'('$stream'('$inoutstream'(StdOut,StdIn))),E0,E) :-
        derefAll(RCmdString,CmdString),
	string2Atom(CmdString,Cmd),
	execCommand(Cmd,StdIn,StdOut,std), E0=E.



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% primitives implementing IORefs:

% New IORefs are represented as mutable values. The "share" constructor
% is put around to be conform with the remaining implementation where
% all mutables are "marked" by this constructor.
?- block 'prim_newIORef'(?,?,?,-,?).
'prim_newIORef'(Val,_,'$io'('IOExts.IORef'(share(MutVal))),E0,E) :-
        var(Val), !,
	create_mutable('$eval'(Val),MutVal), E0=E.
'prim_newIORef'(Val,_,'$io'('IOExts.IORef'(share(MutVal))),E0,E) :-
	create_mutable(Val,MutVal), E0=E.

% When an IORef is read and its value is not evaluated, the current value
% is wrapped into a new mutable in order to implement sharing
% of evaluations of IORefs. The current IORef is updated so that
% it refers to the new mutable (without this indirection, there is
% a risk of creating cyclic structures when the IORef itself is updated).
?- block prim_readIORef(?,?,?,-,?).
prim_readIORef(IORef,_,R,E0,E) :-
        hnf(IORef,HIORef,E0,E1),
	prim_readIORef_exec(HIORef,R,E1,E).

?- block prim_readIORef_exec(-,?,?,?), prim_readIORef_exec(?,?,-,?).
prim_readIORef_exec('IOExts.IORef'(share(MutVal)),'$io'(V),E0,E) :-
        get_mutable(Val,MutVal),
	(Val='$eval'(V) -> true
	  ; create_mutable(Val,MutV),
	    update_mutable(share(MutV),MutVal),
	    V=share(MutV)),
        E0=E.

% Assign a new value to an IORef:
?- block prim_writeIORef(?,?,?,?,-,?).
prim_writeIORef(IORef,Val,_,R,E0,E) :-
        hnf(IORef,HIORef,E0,E1),
	prim_writeIORef_exec(HIORef,Val,R,E1,E).

?- block prim_writeIORef_exec(-,?,?,?,?), prim_writeIORef_exec(?,?,?,-,?).
prim_writeIORef_exec('IOExts.IORef'(share(MutVal)),Val,
		     '$io'('Prelude.()'),E0,E) :-
        var(Val), !,
	update_mutable('$eval'(Val),MutVal), E0=E.
prim_writeIORef_exec('IOExts.IORef'(share(MutVal)),Val,
		     '$io'('Prelude.()'),E0,E) :-
	update_mutable(Val,MutVal), E0=E.

